@@ -3,8 +3,6 @@
 #include <llvm/IR/DataLayout.h>
 
 #include "julia.h"
-#include "julia_internal.h"
-#include "processor.h"
 
 #include <map>
 #include <mutex>
