@@ -38,20 +38,18 @@ STATISTIC(LoweredWithoutFMA, "Number of have_fma's that were lowered to false");
 extern JuliaOJIT *jl_ExecutionEngine;
 
 // whether this platform unconditionally (i.e. without needing multiversioning) supports FMA
-Optional<bool> always_have_fma(Function &intr) {
-    auto intr_name = intr.getName();
-    auto typ = intr_name.substr(strlen("julia.cpu.have_fma."));
-
-#if defined(_CPU_AARCH64_)
-    return typ == "f32" || typ == "f64";
-#else
-    (void)typ;
-    return {};
-#endif
+Optional<bool> always_have_fma(Function &intr, const Triple &TT) JL_NOTSAFEPOINT {
+    if (TT.isAArch64()) {
+        auto intr_name = intr.getName();
+        auto typ = intr_name.substr(strlen("julia.cpu.have_fma."));
+        return typ == "f32" || typ == "f64";
+    } else {
+        return {};
+    }
 }
 
-bool have_fma(Function &intr, Function &caller) {
-    auto unconditional = always_have_fma(intr);
+static bool have_fma(Function &intr, Function &caller, const Triple &TT) JL_NOTSAFEPOINT {
+    auto unconditional = always_have_fma(intr, TT);
     if (unconditional.hasValue())
         return unconditional.getValue();
 
@@ -62,24 +60,24 @@ bool have_fma(Function &intr, Function &caller) {
     StringRef FS =
         FSAttr.isValid() ? FSAttr.getValueAsString() : jl_ExecutionEngine->getTargetFeatureString();
 
-    SmallVector<StringRef, 6> Features;
+    SmallVector<StringRef, 128> Features;
     FS.split(Features, ',');
     for (StringRef Feature : Features)
-#if defined _CPU_ARM_
+    if (TT.isARM()) {
       if (Feature == "+vfp4")
         return typ == "f32" || typ == "f64";
       else if (Feature == "+vfp4sp")
         return typ == "f32";
-#else
+    } else if (TT.isX86()) {
       if (Feature == "+fma" || Feature == "+fma4")
         return typ == "f32" || typ == "f64";
-#endif
+    }
 
     return false;
 }
 
-void lowerHaveFMA(Function &intr, Function &caller, CallInst *I) {
-    if (have_fma(intr, caller)) {
+void lowerHaveFMA(Function &intr, Function &caller, const Triple &TT, CallInst *I) JL_NOTSAFEPOINT {
+    if (have_fma(intr, caller, TT)) {
         ++LoweredWithFMA;
         I->replaceAllUsesWith(ConstantInt::get(I->getType(), 1));
     } else {
@@ -89,8 +87,9 @@ void lowerHaveFMA(Function &intr, Function &caller, CallInst *I) {
     return;
 }
 
-bool lowerCPUFeatures(Module &M)
+bool lowerCPUFeatures(Module &M) JL_NOTSAFEPOINT
 {
+    auto TT = Triple(M.getTargetTriple());
     SmallVector<Instruction*,6> Materialized;
 
     for (auto &F: M.functions()) {
@@ -100,7 +99,7 @@ bool lowerCPUFeatures(Module &M)
             for (Use &U: F.uses()) {
                 User *RU = U.getUser();
                 CallInst *I = cast<CallInst>(RU);
-                lowerHaveFMA(F, *I->getParent()->getParent(), I);
+                lowerHaveFMA(F, *I->getParent()->getParent(), TT, I);
                 Materialized.push_back(I);
             }
         }
@@ -119,7 +118,7 @@ bool lowerCPUFeatures(Module &M)
     }
 }
 
-PreservedAnalyses CPUFeatures::run(Module &M, ModuleAnalysisManager &AM)
+PreservedAnalyses CPUFeaturesPass::run(Module &M, ModuleAnalysisManager &AM)
 {
     if (lowerCPUFeatures(M)) {
         return PreservedAnalyses::allInSet<CFGAnalyses>();
@@ -130,7 +129,7 @@ PreservedAnalyses CPUFeatures::run(Module &M, ModuleAnalysisManager &AM)
 namespace {
 struct CPUFeaturesLegacy : public ModulePass {
     static char ID;
-    CPUFeaturesLegacy() : ModulePass(ID) {};
+    CPUFeaturesLegacy() JL_NOTSAFEPOINT : ModulePass(ID) {};
 
     bool runOnModule(Module &M)
     {
@@ -151,7 +150,8 @@ Pass *createCPUFeaturesPass()
     return new CPUFeaturesLegacy();
 }
 
-extern "C" JL_DLLEXPORT void LLVMExtraAddCPUFeaturesPass_impl(LLVMPassManagerRef PM)
+extern "C" JL_DLLEXPORT_CODEGEN
+void LLVMExtraAddCPUFeaturesPass_impl(LLVMPassManagerRef PM)
 {
     unwrap(PM)->add(createCPUFeaturesPass());
 }
