@@ -12,7 +12,7 @@
 #include <llvm/IR/Module.h>
 #include <llvm/IR/Type.h>
 
-#include "codegen_shared.h"
+#include "llvm-codegen-shared.h"
 #include "julia_assert.h"
 #include "llvm-pass-helpers.h"
 
@@ -27,8 +27,7 @@ JuliaPassContext::JuliaPassContext()
         gc_preserve_begin_func(nullptr), gc_preserve_end_func(nullptr),
         pointer_from_objref_func(nullptr), alloc_obj_func(nullptr),
         typeof_func(nullptr), write_barrier_func(nullptr),
-        write_barrier_binding_func(nullptr), call_func(nullptr),
-        call2_func(nullptr), module(nullptr)
+        call_func(nullptr), call2_func(nullptr), module(nullptr)
 {
 }
 
@@ -51,7 +50,6 @@ void JuliaPassContext::initFunctions(Module &M)
     pointer_from_objref_func = M.getFunction("julia.pointer_from_objref");
     typeof_func = M.getFunction("julia.typeof");
     write_barrier_func = M.getFunction("julia.write_barrier");
-    write_barrier_binding_func = M.getFunction("julia.write_barrier_binding");
     alloc_obj_func = M.getFunction("julia.gc_alloc_obj");
     call_func = M.getFunction("julia.call");
     call2_func = M.getFunction("julia.call2");
@@ -103,7 +101,8 @@ llvm::Function *JuliaPassContext::getOrDeclare(
     else {
         // Otherwise, we'll declare it and add it to the module.
         // Declare the function.
-        auto func = desc.declare(*this);
+        auto T_size = module->getDataLayout().getIntPtrType(module->getContext());
+        auto func = desc.declare(T_size);
         // Add it to the function list.
         module->getFunctionList().push_back(func);
         // Return the newly created function.
@@ -118,26 +117,27 @@ namespace jl_intrinsics {
     static const char *PUSH_GC_FRAME_NAME = "julia.push_gc_frame";
     static const char *POP_GC_FRAME_NAME = "julia.pop_gc_frame";
     static const char *QUEUE_GC_ROOT_NAME = "julia.queue_gc_root";
-    static const char *QUEUE_GC_BINDING_NAME = "julia.queue_gc_binding";
+    static const char *SAFEPOINT_NAME = "julia.safepoint";
 
     // Annotates a function with attributes suitable for GC allocation
     // functions. Specifically, the return value is marked noalias and nonnull.
     // The allocation size is set to the first argument.
-    static Function *addGCAllocAttributes(Function *target, LLVMContext &context)
+    static Function *addGCAllocAttributes(Function *target)
     {
         addRetAttr(target, Attribute::NoAlias);
         addRetAttr(target, Attribute::NonNull);
-        target->addFnAttr(Attribute::getWithAllocSizeArgs(context, 1, None)); // returns %1 bytes
         return target;
     }
 
     const IntrinsicDescription getGCFrameSlot(
         GET_GC_FRAME_SLOT_NAME,
-        [](const JuliaPassContext &context) {
+        [](Type *T_size) {
+            auto &ctx = T_size->getContext();
+            auto T_pprjlvalue = JuliaType::get_pprjlvalue_ty(ctx);
             return Function::Create(
                 FunctionType::get(
-                    PointerType::get(context.T_prjlvalue, 0),
-                    {PointerType::get(context.T_prjlvalue, 0), Type::getInt32Ty(context.getLLVMContext())},
+                    T_pprjlvalue,
+                    {T_pprjlvalue, Type::getInt32Ty(ctx)},
                     false),
                 Function::ExternalLinkage,
                 GET_GC_FRAME_SLOT_NAME);
@@ -145,26 +145,27 @@ namespace jl_intrinsics {
 
     const IntrinsicDescription GCAllocBytes(
         GC_ALLOC_BYTES_NAME,
-        [](const JuliaPassContext &context) {
+        [](Type *T_size) {
+            auto &ctx = T_size->getContext();
+            auto T_prjlvalue = JuliaType::get_prjlvalue_ty(ctx);
             auto intrinsic = Function::Create(
                 FunctionType::get(
-                    context.T_prjlvalue,
-                    { Type::getInt8PtrTy(context.getLLVMContext()),
-                        sizeof(size_t) == sizeof(uint32_t) ?
-                        Type::getInt32Ty(context.getLLVMContext()) :
-                        Type::getInt64Ty(context.getLLVMContext()) },
+                    T_prjlvalue,
+                    { Type::getInt8PtrTy(ctx), T_size },
                     false),
                 Function::ExternalLinkage,
                 GC_ALLOC_BYTES_NAME);
-
-            return addGCAllocAttributes(intrinsic, context.getLLVMContext());
+            intrinsic->addFnAttr(Attribute::getWithAllocSizeArgs(ctx, 1, None));
+            return addGCAllocAttributes(intrinsic);
         });
 
     const IntrinsicDescription newGCFrame(
         NEW_GC_FRAME_NAME,
-        [](const JuliaPassContext &context) {
+        [](Type *T_size) {
+            auto &ctx = T_size->getContext();
+            auto T_pprjlvalue = JuliaType::get_pprjlvalue_ty(ctx);
             auto intrinsic = Function::Create(
-                FunctionType::get(PointerType::get(context.T_prjlvalue, 0), {Type::getInt32Ty(context.getLLVMContext())}, false),
+                FunctionType::get(T_pprjlvalue, {Type::getInt32Ty(ctx)}, false),
                 Function::ExternalLinkage,
                 NEW_GC_FRAME_NAME);
             addRetAttr(intrinsic, Attribute::NoAlias);
@@ -175,11 +176,13 @@ namespace jl_intrinsics {
 
     const IntrinsicDescription pushGCFrame(
         PUSH_GC_FRAME_NAME,
-        [](const JuliaPassContext &context) {
+        [](Type *T_size) {
+            auto &ctx = T_size->getContext();
+            auto T_pprjlvalue = JuliaType::get_pprjlvalue_ty(ctx);
             return Function::Create(
                 FunctionType::get(
-                    Type::getVoidTy(context.getLLVMContext()),
-                    {PointerType::get(context.T_prjlvalue, 0), Type::getInt32Ty(context.getLLVMContext())},
+                    Type::getVoidTy(ctx),
+                    {T_pprjlvalue, Type::getInt32Ty(ctx)},
                     false),
                 Function::ExternalLinkage,
                 PUSH_GC_FRAME_NAME);
@@ -187,11 +190,13 @@ namespace jl_intrinsics {
 
     const IntrinsicDescription popGCFrame(
         POP_GC_FRAME_NAME,
-        [](const JuliaPassContext &context) {
+        [](Type *T_size) {
+            auto &ctx = T_size->getContext();
+            auto T_pprjlvalue = JuliaType::get_pprjlvalue_ty(ctx);
             return Function::Create(
                 FunctionType::get(
-                    Type::getVoidTy(context.getLLVMContext()),
-                    {PointerType::get(context.T_prjlvalue, 0)},
+                    Type::getVoidTy(ctx),
+                    {T_pprjlvalue},
                     false),
                 Function::ExternalLinkage,
                 POP_GC_FRAME_NAME);
@@ -199,11 +204,13 @@ namespace jl_intrinsics {
 
     const IntrinsicDescription queueGCRoot(
         QUEUE_GC_ROOT_NAME,
-        [](const JuliaPassContext &context) {
+        [](Type *T_size) {
+            auto &ctx = T_size->getContext();
+            auto T_prjlvalue = JuliaType::get_prjlvalue_ty(ctx);
             auto intrinsic = Function::Create(
                 FunctionType::get(
-                    Type::getVoidTy(context.getLLVMContext()),
-                    { context.T_prjlvalue },
+                    Type::getVoidTy(ctx),
+                    { T_prjlvalue },
                     false),
                 Function::ExternalLinkage,
                 QUEUE_GC_ROOT_NAME);
@@ -211,16 +218,18 @@ namespace jl_intrinsics {
             return intrinsic;
         });
 
-    const IntrinsicDescription queueGCBinding(
-        QUEUE_GC_BINDING_NAME,
-        [](const JuliaPassContext &context) {
+    const IntrinsicDescription safepoint(
+        SAFEPOINT_NAME,
+        [](Type *T_size) {
+            auto &ctx = T_size->getContext();
+            auto T_psize = T_size->getPointerTo();
             auto intrinsic = Function::Create(
                 FunctionType::get(
-                    Type::getVoidTy(context.getLLVMContext()),
-                    { context.T_prjlvalue },
+                    Type::getVoidTy(ctx),
+                    {T_psize},
                     false),
                 Function::ExternalLinkage,
-                QUEUE_GC_BINDING_NAME);
+                SAFEPOINT_NAME);
             intrinsic->addFnAttr(Attribute::InaccessibleMemOrArgMemOnly);
             return intrinsic;
         });
@@ -230,63 +239,51 @@ namespace jl_well_known {
     static const char *GC_BIG_ALLOC_NAME = XSTR(jl_gc_big_alloc);
     static const char *GC_POOL_ALLOC_NAME = XSTR(jl_gc_pool_alloc);
     static const char *GC_QUEUE_ROOT_NAME = XSTR(jl_gc_queue_root);
-    static const char *GC_QUEUE_BINDING_NAME = XSTR(jl_gc_queue_binding);
     static const char *GC_ALLOC_TYPED_NAME = XSTR(jl_gc_alloc_typed);
 
     using jl_intrinsics::addGCAllocAttributes;
 
     const WellKnownFunctionDescription GCBigAlloc(
         GC_BIG_ALLOC_NAME,
-        [](const JuliaPassContext &context) {
+        [](Type *T_size) {
+            auto &ctx = T_size->getContext();
+            auto T_prjlvalue = JuliaType::get_prjlvalue_ty(ctx);
             auto bigAllocFunc = Function::Create(
                 FunctionType::get(
-                    context.T_prjlvalue,
-                    { Type::getInt8PtrTy(context.getLLVMContext()),
-                        sizeof(size_t) == sizeof(uint32_t) ?
-                        Type::getInt32Ty(context.getLLVMContext()) :
-                        Type::getInt64Ty(context.getLLVMContext()) },
+                    T_prjlvalue,
+                    { Type::getInt8PtrTy(ctx), T_size },
                     false),
                 Function::ExternalLinkage,
                 GC_BIG_ALLOC_NAME);
-
-            return addGCAllocAttributes(bigAllocFunc, context.getLLVMContext());
+            bigAllocFunc->addFnAttr(Attribute::getWithAllocSizeArgs(ctx, 1, None));
+            return addGCAllocAttributes(bigAllocFunc);
         });
 
     const WellKnownFunctionDescription GCPoolAlloc(
         GC_POOL_ALLOC_NAME,
-        [](const JuliaPassContext &context) {
+        [](Type *T_size) {
+            auto &ctx = T_size->getContext();
+            auto T_prjlvalue = JuliaType::get_prjlvalue_ty(ctx);
             auto poolAllocFunc = Function::Create(
                 FunctionType::get(
-                    context.T_prjlvalue,
-                    { Type::getInt8PtrTy(context.getLLVMContext()), Type::getInt32Ty(context.getLLVMContext()), Type::getInt32Ty(context.getLLVMContext()) },
+                    T_prjlvalue,
+                    { Type::getInt8PtrTy(ctx), Type::getInt32Ty(ctx), Type::getInt32Ty(ctx) },
                     false),
                 Function::ExternalLinkage,
                 GC_POOL_ALLOC_NAME);
-
-            return addGCAllocAttributes(poolAllocFunc, context.getLLVMContext());
-        });
-
-    const WellKnownFunctionDescription GCQueueBinding(
-        GC_QUEUE_BINDING_NAME,
-        [](const JuliaPassContext &context) {
-            auto func = Function::Create(
-                FunctionType::get(
-                    Type::getVoidTy(context.getLLVMContext()),
-                    { context.T_prjlvalue },
-                    false),
-                Function::ExternalLinkage,
-                GC_QUEUE_BINDING_NAME);
-            func->addFnAttr(Attribute::InaccessibleMemOrArgMemOnly);
-            return func;
+            poolAllocFunc->addFnAttr(Attribute::getWithAllocSizeArgs(ctx, 2, None));
+            return addGCAllocAttributes(poolAllocFunc);
         });
 
     const WellKnownFunctionDescription GCQueueRoot(
         GC_QUEUE_ROOT_NAME,
-        [](const JuliaPassContext &context) {
+        [](Type *T_size) {
+            auto &ctx = T_size->getContext();
+            auto T_prjlvalue = JuliaType::get_prjlvalue_ty(ctx);
             auto func = Function::Create(
                 FunctionType::get(
-                    Type::getVoidTy(context.getLLVMContext()),
-                    { context.T_prjlvalue },
+                    Type::getVoidTy(ctx),
+                    { T_prjlvalue },
                     false),
                 Function::ExternalLinkage,
                 GC_QUEUE_ROOT_NAME);
@@ -296,19 +293,19 @@ namespace jl_well_known {
 
     const WellKnownFunctionDescription GCAllocTyped(
         GC_ALLOC_TYPED_NAME,
-        [](const JuliaPassContext &context) {
+        [](Type *T_size) {
+            auto &ctx = T_size->getContext();
+            auto T_prjlvalue = JuliaType::get_prjlvalue_ty(ctx);
             auto allocTypedFunc = Function::Create(
                 FunctionType::get(
-                    context.T_prjlvalue,
-                    { Type::getInt8PtrTy(context.getLLVMContext()),
-                        sizeof(size_t) == sizeof(uint32_t) ?
-                        Type::getInt32Ty(context.getLLVMContext()) :
-                        Type::getInt64Ty(context.getLLVMContext()),
-                        Type::getInt8PtrTy(context.getLLVMContext()) },
+                    T_prjlvalue,
+                    { Type::getInt8PtrTy(ctx),
+                        T_size,
+                        Type::getInt8PtrTy(ctx) },
                     false),
                 Function::ExternalLinkage,
                 GC_ALLOC_TYPED_NAME);
-            allocTypedFunc->addFnAttr(Attribute::getWithAllocSizeArgs(context.getLLVMContext(), 1, None));
-            return addGCAllocAttributes(allocTypedFunc, context.getLLVMContext());
+            allocTypedFunc->addFnAttr(Attribute::getWithAllocSizeArgs(ctx, 1, None));
+            return addGCAllocAttributes(allocTypedFunc);
         });
 }
